import textwrap
from typing import Any, Callable

import pytest
from kubernetes import config
from kubernetes.client import CoreV1Api
from kubernetes.client.configuration import Configuration
from kubernetes.client.models import (
    V1Container,
    V1EmptyDirVolumeSource,
    V1ObjectMeta,
    V1Pod,
    V1PodSpec,
    V1Volume,
)


def kubernetes_enabled() -> bool:
    try:
        config.load_config()  # type: ignore
        api_client = CoreV1Api()
        api_client.list_namespace()
        return True
    except Exception:
        return False


@pytest.fixture
def kubeconfig() -> Any:
    config.load_config()  # type: ignore
    return Configuration.get_default_copy()  # type: ignore


@pytest.fixture
def pod_spec() -> Callable[[str], dict[str, Any]]:
    def default_pod_spec(name: str = "test") -> dict[str, Any]:
        return {
            "apiVersion": "v1",
            "kind": "Pod",
            "metadata": {
                "annotations": {"test": "true"},
                "name": name,
                "namespace": "default",
            },
            "spec": {
                "containers": [
                    {
                        "image": "python:3-slim",
                        "name": "autogen-executor",
                        "args": [
                            "/bin/sh",
                            "-c",
                            "echo 'test container'; while true;do sleep 5; done",
                        ],
                    },
                ],
            },
        }

    return default_pod_spec


@pytest.fixture
def pod_yaml_str() -> Callable[[str], str]:
    def default_pod_yaml(name: str = "test") -> str:
        return (
            "apiVersion: v1\n"
            "kind: Pod\n"
            "metadata:\n"
            "  annotations:\n"
            "    test: 'true'\n"
            f"  name: {name}\n"
            "  namespace: default\n"
            "spec:\n"
            "  containers:\n"
            "  - args:\n"
            "    - sh\n"
            "    - -c\n"
            "    - while true;do sleep 5; done\n"
            "    image: python:3-slim\n"
            "    name: autogen-executor\n"
        )

    return default_pod_yaml


@pytest.fixture
def v1_pod() -> Callable[[str, str], Any]:
    def default_v1_pod(name: str = "test", container_name: str = "autogen-executor") -> Any:
        return V1Pod(
            metadata=V1ObjectMeta(name=name, namespace="default"),
            spec=V1PodSpec(
                restart_policy="Never",
                containers=[
                    V1Container(
                        args=[
                            "-c",
                            "echo 'test container'; while true;do sleep 5; done",
                        ],
                        command=["/bin/sh"],
                        name=container_name,
                        image="python:3-slim",
                    )
                ],
            ),
        )

    return default_v1_pod


@pytest.fixture
def volume_dict() -> Callable[[str], dict[str, Any]]:
    def default_volume_dict(name: str = "test") -> dict[str, Any]:
        return {"name": name, "emptyDir": {"medium": "Memory", "sizeLimit": "5Mi"}}

    return default_volume_dict


@pytest.fixture
def v1_volume() -> Callable[[str], Any]:
    def default_v1_volume(name: str = "test") -> Any:
        return V1Volume(
            name=name,
            empty_dir=V1EmptyDirVolumeSource(medium="Memory", size_limit="5Mi"),
        )

    return default_v1_volume


@pytest.fixture
def generated_pod_name_regex() -> str:
    return r"^autogen-code-exec-[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-4[0-9a-fA-F]{3}-[89abAB][0-9a-fA-F]{3}-[0-9a-fA-F]{12}$"



from typing import Any, Callable

import pytest
from autogen_kubernetes.code_executors._utils import (
    StreamChannel,
    clean_none_value,
    create_pod,
    delete_pod,
    get_pod_logs,
    pod_exec_stream,
    wait_for_ready,
)
from kubernetes import config
from kubernetes.client import CoreV1Api
from kubernetes.client.models import (
    V1PersistentVolumeClaimVolumeSource,
    V1Volume,
)


def kubernetes_enabled() -> bool:
    try:
        config.load_config()  # type: ignore
        api_client = CoreV1Api()
        api_client.list_namespace()
        return True
    except Exception:
        return False


state_kubernetes_enabled = kubernetes_enabled()


def test_clean_none_value() -> None:
    volume = V1Volume(
        name="test",
        persistent_volume_claim=V1PersistentVolumeClaimVolumeSource(claim_name="test-pvc", read_only=False),
    )
    volume_dict = clean_none_value(dict(volume.to_dict()))
    assert "aws_elastic_block_store" not in volume_dict
    assert volume_dict["name"] == "test"
    assert volume_dict["persistent_volume_claim"] == {
        "claim_name": "test-pvc",
        "read_only": False,
    }


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_create_pod(kubeconfig: Any, pod_spec: Callable[[str], dict[str, Any]]) -> None:
    create_pod_spec = pod_spec("test-create")
    pod = await create_pod(kubeconfig, create_pod_spec, dry_run=True)

    assert pod["metadata"]["name"] == "test-create"
    assert len(pod["spec"]["containers"]) == 1


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_delete_pod(kubeconfig: Any, pod_spec: Callable[[str], dict[str, Any]]) -> None:
    delete_pod_spec = pod_spec("test-delete")
    pod = await create_pod(kubeconfig, delete_pod_spec)
    deleted_pod = await delete_pod(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
    )

    assert deleted_pod["metadata"]["name"] == "test-delete"


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_get_pod_log(kubeconfig: Any, pod_spec: Callable[[str], dict[str, Any]]) -> None:
    get_log_pod_spec = pod_spec("test-get-log")
    pod = await create_pod(kubeconfig, get_log_pod_spec)
    pod = await wait_for_ready(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
    )
    log = await get_pod_logs(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
        container_name="autogen-executor",
    )
    await delete_pod(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
    )

    assert "test container" in log


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_exec_pod(kubeconfig: Any, pod_spec: Callable[[str], dict[str, Any]]) -> None:
    exec_pod_spec = pod_spec("test-exec")
    pod = await create_pod(kubeconfig, exec_pod_spec)
    pod = await wait_for_ready(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
    )
    command = ["sh", "-c", "python3 - << EOF\nprint('stdout')\nimport sys\nprint('stderr', file=sys.stderr)\nEOF"]
    last_exit_code = 0
    async for channel, msg, exit_code in pod_exec_stream(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
        command=command,
        container_name="autogen-executor",
    ):
        if channel == StreamChannel.STDOUT_CHANNEL:
            assert "stdout" in msg
        elif channel == StreamChannel.STDERR_CHANNEL:
            assert "stderr" in msg
        if exit_code is not None:
            last_exit_code = exit_code

    assert last_exit_code == 0

    await delete_pod(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
    )


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_exec_pod_timeout(kubeconfig: Any, pod_spec: Callable[[str], dict[str, Any]]) -> None:
    exec_pod_spec = pod_spec("test-exec-timeout")
    pod = await create_pod(kubeconfig, exec_pod_spec)
    pod = await wait_for_ready(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
    )
    command = ["sh", "-c", "timeout 3 sleep 5"]
    last_exit_code = 0
    async for _, _, exit_code in pod_exec_stream(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
        command=command,
        container_name="autogen-executor",
    ):
        if exit_code is not None:
            last_exit_code = exit_code

    assert last_exit_code == 124

    await delete_pod(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
    )


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_exec_pod_error(kubeconfig: Any, pod_spec: Callable[[str], dict[str, Any]]) -> None:
    exec_pod_spec = pod_spec("test-exec-error")
    pod = await create_pod(kubeconfig, exec_pod_spec)
    pod = await wait_for_ready(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
    )
    command = ["sh", "-c", "python3 - << EOF\nraise ValueError('test error')\nEOF"]
    last_exit_code = 0
    async for _, _, exit_code in pod_exec_stream(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
        command=command,
        container_name="autogen-executor",
    ):
        if exit_code is not None:
            last_exit_code = exit_code

    assert last_exit_code != 0

    await delete_pod(
        kubeconfig,
        pod_name=pod["metadata"]["name"],
        namespace=pod["metadata"]["namespace"],
    )



from __future__ import annotations

import asyncio
import functools
import inspect
import json
import logging
import os
import re
import ssl
import urllib.parse
from dataclasses import dataclass, field
from enum import IntEnum
from importlib.abc import SourceLoader
from importlib.util import module_from_spec, spec_from_loader
from pathlib import Path
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Generic,
    Optional,
    Sequence,
    Set,
    Type,
    TypeVar,
    Union,
)

import httpx
import kubernetes.client
import yaml
from autogen_core.code_executor import CodeResult
from typing_extensions import ParamSpec
from websockets.asyncio.client import connect
from websockets.exceptions import ConnectionClosed
from websockets.typing import Subprotocol

T = TypeVar("T")
P = ParamSpec("P")


class StreamChannel(IntEnum):
    STDIN_CHANNEL = 0
    STDOUT_CHANNEL = 1
    STDERR_CHANNEL = 2
    ERROR_CHANNEL = 3
    RESIZE_CHANNEL = 4


class HttpStatusCode(IntEnum):
    OK = 200
    CREATED = 201
    ACCEPTED = 202


POD_NAME_PATTERN = r"^[a-z0-9](?:[a-z0-9-]{0,61})[a-z0-9]?$"


def clean_none_value(data: Any) -> Any:
    if isinstance(data, dict):
        return {k: clean_none_value(v) for k, v in data.items() if v is not None}
    elif isinstance(data, list):
        return [clean_none_value(x) for x in data if x is not None]
    else:
        return data


async def wait_for_ready(
    kube_config: Any,
    pod_name: str,
    namespace: str,
    timeout: int = 60,
    stop_time: float = 0.1,
) -> Any:
    elapsed_time = 0.0
    while True:
        pod = await _read_pod_status(kube_config, pod_name, namespace)
        if pod["status"]["phase"] == "Running" or elapsed_time >= timeout:
            break
        await asyncio.sleep(stop_time)
        elapsed_time += stop_time
    if pod["status"]["phase"] != "Running":
        raise ValueError("Pod failed to start")
    else:
        return pod


def _create_ssl_context_and_headers(
    kube_config: Any,
) -> tuple[ssl.SSLContext, dict[str, str]]:
    ca_cert = kube_config.ssl_ca_cert
    ssl_context = ssl.create_default_context(cafile=ca_cert)
    headers = {}

    if "authorization" in kube_config.api_key:  ## Bearer token
        headers.update({"Authorization": kube_config.api_key["authorization"]})
    else:  ## ssl cafile and keyfile
        ssl_context.load_cert_chain(certfile=kube_config.cert_file, keyfile=kube_config.key_file)
    return ssl_context, headers


websocket_subprotocol_kubernetes_api = os.environ.get("WEBSOCKET_SUBPROTOCOL_KUBERNETES_API", "v4.channel.k8s.io")


async def pod_exec_stream(
    kube_config: Any,
    pod_name: str,
    namespace: str,
    command: list[str],
    container_name: str,
) -> AsyncGenerator[tuple[int, str, Optional[int]], None]:
    api_server_url = kube_config.host

    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    additional_headers = [(key, value) for key, value in headers.items()]

    # websocket
    url = f"{api_server_url}/api/v1/namespaces/{namespace}/pods/{pod_name}/exec"
    params = {
        "container": container_name,
        "stdin": "false",
        "stdout": "true",
        "stderr": "true",
        "tty": "false",
    }
    if not command:
        raise ValueError("command must not be empty list")
    command_query_string = "&".join(f"command={urllib.parse.quote_plus(cmd)}" for cmd in command)
    query_string = "&".join(f"{key}={urllib.parse.quote_plus(value)}" for key, value in params.items() if value)
    websocket_url = f"{url}?{command_query_string}&{query_string}".replace("https://", "wss://")
    subprotocols = [Subprotocol(websocket_subprotocol_kubernetes_api)]

    async with connect(
        websocket_url,
        ssl=ssl_context,
        additional_headers=additional_headers,
        subprotocols=subprotocols,
    ) as ws:
        try:
            while True:
                message = await ws.recv(decode=False)
                if len(message) < 2:
                    continue
                channel = int(message[0])
                msg_content = message[1:]
                if isinstance(msg_content, bytes):
                    content = msg_content.decode("utf-8")
                else:
                    content = msg_content

                returncode = None
                if channel == StreamChannel.ERROR_CHANNEL:  # ERROR_CHANNEL
                    error_info = yaml.safe_load(content)
                    if error_info["status"] == "Success":
                        returncode = 0
                    else:
                        returncode = int(error_info["details"]["causes"][0]["message"])

                yield (channel, content, returncode)

        except ConnectionClosed as e:
            logging.info(f"Websocket connection closed: {e}")


async def _read_pod_status(kube_config: Any, pod_name: str, namespace: str) -> Any:
    api_server_url = kube_config.host
    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    headers.update({"Accept": "application/json"})

    url = f"{api_server_url}/api/v1/namespaces/{namespace}/pods/{pod_name}/status"

    async with httpx.AsyncClient(verify=ssl_context) as httpx_client:
        response = await httpx_client.get(url, headers=headers)
        if response.status_code == HttpStatusCode.OK:
            return response.json()
        else:
            logging.info("Failed to get pod status: {response.status_code}")
            response.raise_for_status()


async def get_pod_logs(
    kube_config: Any,
    pod_name: str,
    namespace: str,
    container_name: str,
) -> Any:
    api_server_url = kube_config.host
    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    headers.update({"Accept": "application/json", "Content-Type": "application/json"})

    url = f"{api_server_url}/api/v1/namespaces/{namespace}/pods/{pod_name}/log?container={container_name}"

    async with httpx.AsyncClient(verify=ssl_context) as httpx_client:
        response = await httpx_client.get(url, headers=headers)
        if response.status_code == HttpStatusCode.OK:
            return response.text
        else:
            logging.info("Failed to get pod status: {response.status_code}")
            response.raise_for_status()


async def create_pod(
    kube_config: Any,
    pod_spec: dict[str, Any],
    dry_run: bool = False,
) -> Any:
    api_server_url = kube_config.host
    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    headers.update(
        {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
    )

    namespace = pod_spec["metadata"]["namespace"]
    url = f"{api_server_url}/api/v1/namespaces/{namespace}/pods"
    if dry_run:
        url += "?dryRun=All"

    async with httpx.AsyncClient(verify=ssl_context) as httpx_client:
        response = await httpx_client.post(url, headers=headers, json=pod_spec)
        if response.status_code == HttpStatusCode.CREATED:
            return response.json()
        else:
            logging.info("Failed to create pod: {response.status_code}")
            response.raise_for_status()


async def delete_pod(kube_config: Any, pod_name: str, namespace: str) -> Any:
    api_server_url = kube_config.host
    ssl_context, headers = _create_ssl_context_and_headers(kube_config)
    headers.update({"Accept": "application/json"})

    url = f"{api_server_url}/api/v1/namespaces/{namespace}/pods/{pod_name}"

    async with httpx.AsyncClient(verify=ssl_context) as httpx_client:
        response = await httpx_client.delete(url, headers=headers)
        if response.status_code in [HttpStatusCode.OK, HttpStatusCode.ACCEPTED]:
            return response.json()
        else:
            logging.info("Failed to get pod status: {response.status_code}")
            response.raise_for_status()


# Source below based from: https://github.com/microsoft/autogen/blob/main/python/packages/autogen-ext/src/autogen_ext/code_executors/_common.py
# Credit to original authors
# Original code Licensed under the MIT License.
# See the License file for the full license text.


@dataclass
class CommandLineCodeResult(CodeResult):
    """A code result class for command line code executor."""

    code_file: Optional[str]


# Source below based from: https://github.com/microsoft/autogen/blob/main/python/packages/autogen-core/src/autogen_core/code_executor/_func_with_reqs.py
# Credit to original authors
# Original code Licensed under the MIT License.
# Modifications made by questcollector
# See the License file for the full license text.


@dataclass(frozen=True)
class Alias:
    name: str
    alias: str


# make hashable
@dataclass(frozen=True)
class ImportFromModule:
    module: str
    imports: tuple[Union[str, Alias]]


Import = Union[str, ImportFromModule, Alias]


class _StringLoader(SourceLoader):
    def __init__(self, data: str):
        self.data = data

    def get_source(self, fullname: str) -> str:
        return self.data

    def get_data(self, path: str) -> bytes:
        return self.data.encode("utf-8")

    def get_filename(self, fullname: str) -> str:
        return "<not a real path>/" + fullname + ".py"


@dataclass
class FunctionWithRequirementsStr:
    func: str
    compiled_func: Callable[..., Any]
    _func_name: str
    python_packages: Sequence[str] = field(default_factory=list)
    global_imports: Sequence[Import] = field(default_factory=list)

    def __init__(
        self,
        func: str,
        python_packages: Sequence[str] = [],
        global_imports: Sequence[Import] = [],
    ):
        self.func = func
        self.python_packages = python_packages
        self.global_imports = global_imports

        module_name = "func_module"
        loader = _StringLoader(func)
        spec = spec_from_loader(module_name, loader)
        if spec is None:
            raise ValueError("Could not create spec")
        module = module_from_spec(spec)
        if spec.loader is None:
            raise ValueError("Could not create loader")

        try:
            spec.loader.exec_module(module)
        except Exception as e:
            raise ValueError(f"Could not compile function: {e}") from e

        functions = inspect.getmembers(module, inspect.isfunction)
        if len(functions) != 1:
            raise ValueError("The string must contain exactly one function")

        self._func_name, self.compiled_func = functions[0]

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError("String based function with requirement objects are not directly callable")


@dataclass
class FunctionWithRequirements(Generic[T, P]):
    func: Callable[P, T]
    python_packages: Sequence[str] = field(default_factory=list)
    global_imports: Sequence[Import] = field(default_factory=list)

    @classmethod
    def from_callable(
        cls,
        func: Callable[P, T],
        python_packages: Sequence[str] = [],
        global_imports: Sequence[Import] = [],
    ) -> FunctionWithRequirements[T, P]:
        return cls(python_packages=python_packages, global_imports=global_imports, func=func)

    @staticmethod
    def from_str(
        func: str,
        python_packages: Sequence[str] = [],
        global_imports: Sequence[Import] = [],
    ) -> FunctionWithRequirementsStr:
        return FunctionWithRequirementsStr(func=func, python_packages=python_packages, global_imports=global_imports)

    # Type this based on F
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self.func(*args, **kwargs)


def _to_code(func: Union[FunctionWithRequirements[T, P], Callable[P, T], FunctionWithRequirementsStr]) -> str:
    if isinstance(func, FunctionWithRequirementsStr):
        return func.func

    code = inspect.getsource(func)
    # Strip the decorator
    # any decorator not supported
    if code.startswith("@"):
        code = code[code.index("def") :]
    return code


def _import_to_str(im: Import) -> str:
    if isinstance(im, str):
        return f"import {im}"
    elif isinstance(im, Alias):
        return f"import {im.name} as {im.alias}"
    else:

        def to_str(i: Union[str, Alias]) -> str:
            if isinstance(i, str):
                return i
            else:
                return f"{i.name} as {i.alias}"

        imports = ", ".join(map(to_str, im.imports))
        return f"from {im.module} import {imports}"


def build_python_functions_file(
    funcs: Sequence[
        Union[
            FunctionWithRequirements[Any, P],
            Callable[..., Any],
            FunctionWithRequirementsStr,
        ]
    ],
) -> str:
    """:meta private:"""
    # First collect all global imports
    # TODO: Import is not hashable
    global_imports: Set[Import] = set()
    for func in funcs:
        if isinstance(func, (FunctionWithRequirements, FunctionWithRequirementsStr)):
            global_imports.update(func.global_imports)

    content = "\n".join(map(_import_to_str, global_imports)) + "\n\n"

    for func in funcs:
        content += _to_code(func) + "\n\n"

    return content


# Raises ValueError if the file is not in the workspace
def get_file_name_from_content(code: str, workspace_path: Path) -> Optional[str]:
    first_line = code.split("\n")[0]
    # TODO - support other languages
    if first_line.startswith("# filename:"):
        filename = first_line.split(":")[1].strip()

        # Handle relative paths in the filename
        path = Path(filename)
        if not path.is_absolute():
            path = workspace_path / path
        path = path.resolve()
        # Throws an error if the file is not in the workspace
        relative = path.relative_to(workspace_path.resolve())
        return str(relative)

    return None


PYTHON_VARIANTS = ["python", "Python", "py"]


def lang_to_cmd(lang: str) -> str:
    if lang in PYTHON_VARIANTS:
        return "python"
    if lang.startswith("python") or lang in ["bash", "sh"]:
        return lang
    if lang in ["shell"]:
        return "sh"
    else:
        raise ValueError(f"Unsupported language: {lang}")


def silence_pip(code: str, lang: str) -> str:
    """Apply -qqq flag to pip install commands."""
    if lang == "python":
        regex = r"^! ?pip install"
    elif lang in ["bash", "shell", "sh", "pwsh", "powershell", "ps1"]:
        regex = r"^pip install"
    else:
        return code

    # Find lines that start with pip install and make sure "-qqq" flag is added.
    lines = code.split("\n")
    for i, line in enumerate(lines):
        # use regex to find lines that start with pip install.
        match = re.search(regex, line)
        if match is not None:
            if "-qqq" not in line:
                lines[i] = line.replace(match.group(0), match.group(0) + " -qqq")
    return "\n".join(lines)


def with_requirements(
    python_packages: Sequence[str] = [], global_imports: Sequence[Import] = []
) -> Callable[[Callable[P, T]], FunctionWithRequirements[T, P]]:
    """Decorate a function with package and import requirements

    Args:
        python_packages (List[str], optional): Packages required to function. Can include version info.. Defaults to [].
        global_imports (List[Import], optional): Required imports. Defaults to [].

    Returns:
        Callable[[Callable[P, T]], FunctionWithRequirements[T, P]]: The decorated function
    """

    def wrapper(func: Callable[P, T]) -> FunctionWithRequirements[T, P]:
        func_with_reqs = FunctionWithRequirements(
            python_packages=python_packages, global_imports=global_imports, func=func
        )

        functools.update_wrapper(func_with_reqs, func)
        return func_with_reqs

    return wrapper



# File based from: https://github.com/microsoft/autogen/blob/main/python/packages/autogen-ext/src/autogen_ext/code_executors/_docker_code_executor.py
# Credit to original authors
# Original code Licensed under the MIT License.
# Modifications made by kiyoung you(questcollector)
# See the License file for the full license text.

from __future__ import annotations

import asyncio
import logging
import re
import shlex
import sys
import time
import uuid
import warnings
from collections.abc import Coroutine, Sequence
from hashlib import sha256
from pathlib import Path
from types import TracebackType
from typing import Any, Callable, ClassVar, Dict, List, Optional, ParamSpec, Type, Union

import kubernetes.client
import kubernetes.config
import yaml
from autogen_core import CancellationToken
from autogen_core.code_executor import (
    CodeBlock,
    CodeExecutor,
)
from httpx import HTTPStatusError

from ._utils import (
    POD_NAME_PATTERN,
    CommandLineCodeResult,
    FunctionWithRequirements,
    FunctionWithRequirementsStr,
    StreamChannel,
    build_python_functions_file,
    clean_none_value,
    create_pod,
    delete_pod,
    get_file_name_from_content,
    get_pod_logs,
    lang_to_cmd,
    pod_exec_stream,
    silence_pip,
    wait_for_ready,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


A = ParamSpec("A")


# TODO autogen compatibility
class PodCommandLineCodeExecutor(CodeExecutor):
    """Executes code through a command line environment in a container on Kubernetes Pod.

    The executor first saves each code block in a file in the working
    directory, and then executes the code file in the container.
    The executor executes the code blocks in the order they are received.
    Currently, the executor only supports Python and shell scripts.
    For Python code, use the language "python" for the code block.
    For shell scripts, use the language "bash", "shell", or "sh" for the code
    block.

    Args:
        image (_type_, optional): container image to use for code execution.
            Defaults to "python:3-slim".
        pod_name (Optional[str], optional): Name of the kubernetes pod
            which is created. If None, will autogenerate a name. Defaults to None.
        timeout (int, optional): The timeout for code execution. Defaults to 60.
        workspace_path (Union[Path, str], optional): The workspace directory for code executor container.
            Generated code script files will be stored in this directory.
            Not supports relative path(use absolute path). Defaults to Path("/workspace")
        namespace (str, optional): Name of the namespace of kubernetes cluster on which code executor pod will be created.
            Defaults to "default".
        volume (Union[dict, str, Path, kubernetes.client.models.V1Volume, None], optional): The volume for the code execution pod.
            Supports the formats of a dictionary, a YAML string, a YAML file path, and kubernetes `V1Volume` model format.
            Must conform to the kubernetes `V1Volume` model format.
            Must have appropriate access mode(such as ReadWriteMany, ReadWriteOnce, ReadWriteOncePod, in case of PersistentVolumeClaim)
            If None, no volume attached to code executor pod. Defaults to None.
        pod_spec (Union[dict, str, Path, kubernetes.client.models.V1Pod, None], optional): pod specification for code executor.
            Must contain a container which name is "autogen-executor" for execution for codes.
            Supports the formats of a dictionary, a YAML string, a YAML file path, and a kubernetes V1Pod model.
            Must conform to the kubernetes `V1Pod` model format.
            If None, will use above parameters to create code executor pod. Defaults to None.
        kube_config_file (Union[Path, str, None], optional): kubernetes configuration file(kubeconfig) path.
            If None, will use KUBECONFIG environment variables or service account token(incluster config).
            Using service account token, service account must have at least those namespaced permissions below.
            [
              {
                "resource": "pods", "verb": ["get", "create", "delete"]
              },
              {
                "resource": "pods/status", "verb": ["get"]
              },
              {
                "resource": "pods/exec", "verb": ["create"]
              },
              {
                "resource": "pods/log", "verb": ["get"]
              }
            ]
        auto_remove (bool, optional): If true, will automatically remove the
            container when remove is called, when the context manager exits or when
            the Python process exits with atext. Defaults to True.
        functions (List[Union[FunctionWithRequirements[Any, A], Callable[..., Any]]]): A list of functions that are available to the code executor. Default is an empty list.
        functions_module (str, optional): The name of the module that will be created to store the functions. Defaults to "functions".
    """

    SUPPORTED_LANGUAGES: ClassVar[List[str]] = [
        "bash",
        "shell",
        "sh",
        "pwsh",
        "powershell",
        "ps1",
        "python",
    ]

    LANGUAGE_FILE_EXTENSION: ClassVar[Dict[str, str]] = {
        "python": "py",
        "bash": "sh",
        "shell": "sh",
        "sh": "sh",
        "pwsh": "ps1",
        "powershell": "ps1",
        "ps1": "ps1",
    }

    FUNCTION_PROMPT_TEMPLATE: ClassVar[
        str
    ] = """You have access to the following user defined functions. They can be accessed from the module called `$module_name` by their function names.

For example, if there was a function called `foo` you could import it by writing `from $module_name import foo`

$functions"""

    def __init__(
        self,
        image: str = "python:3-slim",
        pod_name: Optional[str] = None,
        *,
        timeout: int = 60,
        workspace_path: Union[Path, str] = Path("/workspace"),
        namespace: str = "default",
        volume: Union[dict[str, Any], str, Path, Type[kubernetes.client.models.V1Volume], None] = None,
        pod_spec: Union[dict[str, Any], str, Path, Type[kubernetes.client.models.V1Pod], None] = None,
        kube_config_file: Union[Path, str, None] = None,
        auto_remove: bool = True,
        functions: Sequence[
            Union[
                FunctionWithRequirements[Any, A],
                Callable[..., Any],
                FunctionWithRequirementsStr,
            ]
        ] = [],
        functions_module: str = "functions",
    ):
        if timeout < 1:
            raise ValueError("Timeout must be greater than or equal to 1.")

        ## kubeconfig
        if isinstance(kube_config_file, str):
            kube_config_file = Path(kube_config_file)

        if kube_config_file is None:  ## configuration from default kubeconfig or incluster
            kubernetes.config.load_config()  # type: ignore
        else:
            kubernetes.config.load_config(config_file=kube_config_file)  # type: ignore
        self._kube_config = kubernetes.client.Configuration.get_default_copy()  # type: ignore

        ## workspace
        if isinstance(workspace_path, str):  ## path string to Path
            workspace_path = Path(workspace_path)

        if not workspace_path.is_absolute():  ## validate workspace_path
            raise ValueError("Not supports workspace path as relative path")

        self._workspace_path: Path = workspace_path

        ## volume
        volume_json = None
        if isinstance(volume, str):
            try:
                if Path(volume).exists():  # YAML file path string
                    volume = Path(volume)
            except OSError:
                pass

        if isinstance(volume, Path):  # YAML file to dict
            volume_json = yaml.safe_load(volume.read_text(encoding="utf-8"))
        elif isinstance(volume, str):  # YAML string to dict
            volume_json = yaml.safe_load(volume)
        elif isinstance(volume, kubernetes.client.models.V1Volume):
            volume_json = clean_none_value(dict(volume.to_dict()))
        elif isinstance(volume, dict):
            volume_json = volume

        self._volume = volume_json

        ## pod_name
        if pod_name is None:
            self._pod_name = f"autogen-code-exec-{uuid.uuid4()}"
        else:
            self._pod_name = pod_name

        if not re.fullmatch(POD_NAME_PATTERN, self._pod_name):
            raise ValueError(
                "Pod name validation failed: pod name must start and end with lower alphanumeric characters, "
                "and contain only lowercase alphanumeric or dash('-') characters."
            )

        self._container_name = "autogen-executor"
        self._namespace = namespace
        self._timeout = timeout
        self._auto_remove = auto_remove
        self._image = image

        ## pod_spec
        if isinstance(pod_spec, str):  # YAML file path string
            try:
                if Path(pod_spec).exists():
                    pod_spec = Path(pod_spec)
            except OSError:
                pass

        pod_spec_dict = {}
        if isinstance(pod_spec, Path):  # YAML file to dict
            pod_spec_dict = yaml.safe_load(pod_spec.read_text(encoding="utf-8"))
        elif isinstance(pod_spec, str):  # YAML string to dict
            pod_spec_dict = yaml.safe_load(pod_spec)
        elif isinstance(pod_spec, kubernetes.client.models.V1Pod):
            pod_spec_dict = clean_none_value(dict(pod_spec.to_dict()))
        elif isinstance(pod_spec, dict):
            pod_spec_dict = pod_spec

        if pod_spec is None:  ## create pod_spec from other parameters
            self._pod = self._define_pod_spec()
        else:
            self._pod = pod_spec_dict

        self._validate_pod()

        if not functions_module.isidentifier():
            raise ValueError("Module name must be a valid Python identifier")

        self._functions_module = functions_module
        self._functions = functions
        # Setup could take some time so we intentionally wait for the first code block to do it.
        if len(functions) > 0:
            self._setup_functions_complete = False
        else:
            self._setup_functions_complete = True

        self._running = False

    @property
    def timeout(self) -> int:
        """(Experimental) The timeout for code execution."""
        return self._timeout

    @property
    def workspace_path(self) -> Path:
        """(Experimental) The working directory for the code execution."""
        return self._workspace_path

    def _validate_pod(self) -> None:
        try:
            # pod_name, namespace
            self._pod_name = self._pod["metadata"]["name"]
            self._namespace = self._pod["metadata"]["namespace"]

            # container
            executor_container_flag = True
            for container in self._pod["spec"]["containers"]:
                if container["name"] == "autogen-executor":
                    executor_container_flag = False
            if executor_container_flag:
                raise ValueError("No container which name is autogen-executor")

        except KeyError as e:
            raise ValueError(e) from e
        except ValueError as e:
            raise e from e

    def _define_pod_spec(self) -> Any:
        pod = kubernetes.client.models.V1Pod()
        metadata = kubernetes.client.models.V1ObjectMeta(name=self._pod_name, namespace=self._namespace)

        executor_container = kubernetes.client.models.V1Container(
            args=["-c", "while true;do sleep 5; done"],
            command=["/bin/sh"],
            name=self._container_name,
            image=self._image,
        )
        pod_spec = kubernetes.client.models.V1PodSpec(restart_policy="Never", containers=[executor_container])

        pod.metadata = metadata
        pod.spec = pod_spec

        pod_manifest: Dict[str, Any] = clean_none_value(dict(pod.to_dict()))

        if self._volume is not None:
            # add volume
            volume: Dict[str, Any] = self._volume
            pod_manifest["spec"]["volumes"] = [volume]

            # add volume mounts to executor container
            pod_manifest["spec"]["containers"][0]["volumeMounts"] = [
                {"mountPath": str(self.workspace_path), "name": volume["name"]}
            ]

        return pod_manifest

    async def _setup_functions(self, cancellation_token: CancellationToken) -> None:  # type: ignore
        func_file_content = build_python_functions_file(self._functions)
        func_file = self.workspace_path / f"{self._functions_module}.py"

        write_func_file_command = (
            f'if [ ! -d "{self.workspace_path}" ]; then\n  mkdir {self.workspace_path}\nfi\n'
            f"cat <<EOF >{func_file}\n{func_file_content}\nEOF\nchmod +x {func_file}"
        )
        write_func_file_stderr_msg: List[str] = []
        write_func_file_stdout_msg: List[str] = []
        write_func_file_exit_code = 0
        async for channel, msg, exit_code in pod_exec_stream(
            self._kube_config,
            self._pod["metadata"]["name"],
            self._pod["metadata"]["namespace"],
            ["sh", "-c", write_func_file_command],
            self._container_name,
        ):
            if channel == StreamChannel.STDOUT_CHANNEL:
                write_func_file_stdout_msg.append(msg)
            elif channel == StreamChannel.STDERR_CHANNEL:
                write_func_file_stderr_msg.append(msg)
            if exit_code is not None:
                write_func_file_exit_code = exit_code

        if write_func_file_exit_code != 0:
            write_func_file_stdout = "".join(write_func_file_stdout_msg)
            write_func_file_stderr = "".join(write_func_file_stderr_msg)
            raise ValueError(f"write function module file failed. \n{write_func_file_stdout}\n{write_func_file_stderr}")

        # Collect requirements
        lists_of_packages = [
            x.python_packages
            for x in self._functions
            if isinstance(x, (FunctionWithRequirements, FunctionWithRequirementsStr))
        ]
        flattened_packages = [item for sublist in lists_of_packages for item in sublist]
        required_packages = list(set(flattened_packages))
        if len(required_packages) > 0:
            logging.info("Ensuring packages are installed in executor.")

            packages = shlex.join(required_packages)

            install_packages_stderr_msg: List[str] = []
            install_packages_stdout_msg: List[str] = []
            install_packages_exit_code = 0
            async for channel, msg, exit_code in pod_exec_stream(
                self._kube_config,
                self._pod["metadata"]["name"],
                self._pod["metadata"]["namespace"],
                ["sh", "-c", f"python -m pip install {packages}"],
                self._container_name,
            ):
                if channel == StreamChannel.STDOUT_CHANNEL:
                    install_packages_stdout_msg.append(msg)
                elif channel == StreamChannel.STDERR_CHANNEL:
                    install_packages_stderr_msg.append(msg)
                if exit_code is not None:
                    install_packages_exit_code = exit_code

            if install_packages_exit_code != 0:
                install_packages_stdout = "".join(install_packages_stdout_msg)
                install_packages_stderr = "".join(install_packages_stderr_msg)
                raise ValueError(f"Pip install failed. \n{install_packages_stdout}\n{install_packages_stderr}")

        # Attempt to load the function file to check for syntax errors, imports etc.
        # TODO use exec
        exec_stderr_msg: List[str] = []
        exec_stdout_msg: List[str] = []
        exec_exit_code = 0
        async for channel, msg, exit_code in pod_exec_stream(
            self._kube_config,
            self._pod["metadata"]["name"],
            self._pod["metadata"]["namespace"],
            ["sh", "-c", "python", str(func_file)],
            self._container_name,
        ):
            if channel == StreamChannel.STDOUT_CHANNEL:
                exec_stdout_msg.append(msg)
            elif channel == StreamChannel.STDERR_CHANNEL:
                exec_stderr_msg.append(msg)
            if exit_code is not None:
                exec_exit_code = exit_code

        if exec_exit_code != 0:
            exec_stdout = "".join(exec_stdout_msg)
            exec_stderr = "".join(exec_stderr_msg)
            raise ValueError(f"Functions failed to load: \n{exec_stdout}\n{exec_stderr}")

        self._setup_functions_complete = True

    async def _execute_code_dont_check_setup(
        self, code_blocks: List[CodeBlock], cancellation_token: CancellationToken
    ) -> CommandLineCodeResult:  # type: ignore
        if self._pod is None or not self._running:
            raise ValueError("Container is not running. Must first be started with either start or a context manager.")

        if len(code_blocks) == 0:
            raise ValueError("No code blocks to execute.")

        outputs: List[str] = []
        files: List[Path] = []
        last_exit_code = 0
        for code_block in code_blocks:
            lang = code_block.language.lower()
            code = silence_pip(code_block.code, lang)

            # Check if there is a filename comment
            try:
                filename = get_file_name_from_content(code, self._workspace_path)
            except ValueError:
                outputs.append("Filename is not in the workspace")
                last_exit_code = 1
                break

            if not filename:
                extension = self.LANGUAGE_FILE_EXTENSION[lang]
                filename = f"tmp_code_{sha256(code.encode()).hexdigest()}.{extension}"

            code_path = self._workspace_path / filename
            mkdir_and_write_code = (
                f'if [ ! -d "{self.workspace_path}" ]; then\n  mkdir {self.workspace_path}\nfi\n'
                f"cat <<EOF >{code_path}\n{code}\nEOF\nchmod +x {code_path}"
            )

            code_write_command = ["sh", "-c", mkdir_and_write_code]
            stderr_msg: List[str] = []
            stdout_msg: List[str] = []
            last_exit_code = 0
            async for channel, msg, exit_code in pod_exec_stream(
                self._kube_config,
                self._pod["metadata"]["name"],
                self._pod["metadata"]["namespace"],
                code_write_command,
                self._container_name,
            ):
                if channel == StreamChannel.STDOUT_CHANNEL:
                    stdout_msg.append(msg)
                elif channel == StreamChannel.STDERR_CHANNEL:
                    stderr_msg.append(msg)
                if exit_code is not None:
                    last_exit_code = exit_code

            outputs.extend(stderr_msg)
            if last_exit_code != 0:
                break

            files.append(code_path)
            command = ["timeout", str(self.timeout), lang_to_cmd(lang), str(code_path)]

            stderr_msg = []
            stdout_msg = []
            async for channel, msg, exit_code in pod_exec_stream(
                self._kube_config,
                self._pod["metadata"]["name"],
                self._pod["metadata"]["namespace"],
                command,
                self._container_name,
            ):
                if channel == StreamChannel.STDOUT_CHANNEL:
                    stdout_msg.append(msg)
                elif channel == StreamChannel.STDERR_CHANNEL:
                    stderr_msg.append(msg)
                if exit_code is not None:
                    last_exit_code = exit_code

            if last_exit_code == 124:
                stdout_msg.append("\n Timeout")
            outputs.extend(stdout_msg)
            outputs.extend(stderr_msg)

            if last_exit_code != 0:
                break

        code_file = str(files[0]) if files else None
        return CommandLineCodeResult(exit_code=last_exit_code, output="".join(outputs), code_file=code_file)

    async def execute_code_blocks(
        self, code_blocks: List[CodeBlock], cancellation_token: CancellationToken
    ) -> CommandLineCodeResult:
        """(Experimental) Execute the code blocks and return the result.

        Args:
            code_blocks (List[CodeBlock]): The code blocks to execute.

        Returns:
            CommandlineCodeResult: The result of the code execution."""

        def raise_not_implemented() -> None:
            raise NotImplementedError("Cancellation is not yet supported for PodCommandLineCodeExecutor")

        cancellation_token.add_callback(lambda: raise_not_implemented())

        if not self._setup_functions_complete:
            await self._setup_functions(cancellation_token)

        return await self._execute_code_dont_check_setup(code_blocks, cancellation_token)

    async def remove(self) -> None:
        await delete_pod(
            self._kube_config,
            self._pod["metadata"]["name"],
            self._pod["metadata"]["namespace"],
        )
        self._running = False

    async def start(self) -> None:
        import asyncio_atexit

        self._pod = await create_pod(self._kube_config, self._pod)

        self._pod = await wait_for_ready(
            self._kube_config,
            self._pod["metadata"]["name"],
            self._pod["metadata"]["namespace"],
            self._timeout,
        )

        async def cleanup() -> None:
            await self.remove()
            asyncio_atexit.unregister(cleanup)  # type: ignore

        if self._auto_remove:
            asyncio_atexit.register(cleanup)  # type: ignore

        # Check if the container is running
        if self._pod["status"]["phase"] != "Running":
            logs_str = await get_pod_logs(
                self._kube_config,
                self._pod["metadata"]["name"],
                self._pod["metadata"]["namespace"],
                self._container_name,
            )
            raise ValueError(f"Failed to start container from image {self._image}. Logs: {logs_str}")

        self._running = True

    async def restart(self) -> None:
        warnings.warn(
            "Restarting Pod command line code executor is not supported. No action is taken.",
            stacklevel=2,
        )

    async def __aenter__(self) -> Self:
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        await self.remove()
        return None



from ._kubernetes_code_executor import PodCommandLineCodeExecutor
from ._utils import (
    Alias,
    FunctionWithRequirements,
    FunctionWithRequirementsStr,
    ImportFromModule,
    with_requirements,
)

__all__ = [
    "PodCommandLineCodeExecutor",
    "FunctionWithRequirements",
    "FunctionWithRequirementsStr",
    "Alias",
    "ImportFromModule",
    "with_requirements",
]



import asyncio
import inspect
import logging
import os
import re
from collections.abc import Callable
from pathlib import Path
from typing import Any

import httpx as hx
import pytest
from autogen_agentchat.agents import CodeExecutorAgent
from autogen_agentchat.messages import TextMessage
from autogen_core import CancellationToken
from autogen_core.code_executor import CodeBlock
from autogen_kubernetes.code_executors import PodCommandLineCodeExecutor
from autogen_kubernetes.code_executors._utils import (
    Alias,
    FunctionWithRequirements,
    ImportFromModule,
    with_requirements,
)
from kubernetes.client import CoreV1Api
from kubernetes.config import load_config  # type: ignore

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def kubernetes_enabled() -> bool:
    try:
        load_config()  # type: ignore
        api_client = CoreV1Api()
        api_client.list_namespace()
        return True
    except Exception:
        return False


state_kubernetes_enabled = kubernetes_enabled()


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_pod_spec_yaml_file() -> None:
    yaml_file_path = Path(os.path.dirname(__file__)) / Path("test-pod.yaml")
    async with PodCommandLineCodeExecutor(pod_spec=str(yaml_file_path)) as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        logger.info(code_result)
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_pod_spec_yaml_str(pod_yaml_str: Callable[[str], str]) -> None:
    async with PodCommandLineCodeExecutor(pod_spec=pod_yaml_str("test-pod-yaml-str")) as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_pod_spec_v1_pod(v1_pod: Callable[[str, str], Any]) -> None:
    v1_pod_spec = v1_pod("test-pod-spec-v1-pod", "autogen-executor")
    async with PodCommandLineCodeExecutor(pod_spec=v1_pod_spec) as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_pod_spec_dict(pod_spec: Callable[[str], dict[str, Any]]) -> None:
    pod_spec_dict = pod_spec("test-pod-spec-dict")
    async with PodCommandLineCodeExecutor(pod_spec=pod_spec_dict) as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_pod_default(generated_pod_name_regex: str) -> None:
    async with PodCommandLineCodeExecutor() as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        assert re.fullmatch(generated_pod_name_regex, executor._pod_name) is not None
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_timeout_error() -> None:
    with pytest.raises(ValueError):
        PodCommandLineCodeExecutor(image="python:3-slim", timeout=0)


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_relative_path_error() -> None:
    with pytest.raises(ValueError):
        PodCommandLineCodeExecutor(image="python:3-slim", workspace_path="./workspace")


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_volume_yaml_file(generated_pod_name_regex: str) -> None:
    yaml_file_path = Path(os.path.dirname(__file__)) / Path("test-volume.yaml")
    async with PodCommandLineCodeExecutor(volume=str(yaml_file_path)) as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        assert re.fullmatch(generated_pod_name_regex, executor._pod_name) is not None
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_volume_yaml_str(generated_pod_name_regex: str) -> None:
    yaml_file_path = Path(os.path.dirname(__file__)) / Path("test-volume.yaml")
    async with PodCommandLineCodeExecutor(volume=yaml_file_path.read_text()) as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        assert re.fullmatch(generated_pod_name_regex, executor._pod_name) is not None
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_volume_dict(volume_dict: Callable[[str], dict[str, Any]], generated_pod_name_regex: str) -> None:
    volume = volume_dict("test")
    async with PodCommandLineCodeExecutor(volume=volume) as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        assert re.fullmatch(generated_pod_name_regex, executor._pod_name) is not None
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_volume_V1Volume(v1_volume: Callable[[str], Any], generated_pod_name_regex: str) -> None:
    v1_volume_obj = v1_volume("test")
    async with PodCommandLineCodeExecutor(volume=v1_volume_obj) as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        assert re.fullmatch(generated_pod_name_regex, executor._pod_name) is not None
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_pod_name_error() -> None:
    invalid_pod_name = "!%#$SDAbsdbawpup230bkvbl;kouqw98q;asdnlkjdvnblafjdg8WU0 2Q3Y5OUDAJFNBJADBVNA;DWERY"
    with pytest.raises(ValueError):
        PodCommandLineCodeExecutor(pod_name=invalid_pod_name)


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_invalid_pod_spec(v1_pod: Callable[[str, str], Any]) -> None:
    invalid_v1_pod = v1_pod("test-invalid-pod", "abcd")
    with pytest.raises(ValueError):
        PodCommandLineCodeExecutor(pod_spec=invalid_v1_pod)


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_func_modules(generated_pod_name_regex: str) -> None:
    test_function = FunctionWithRequirements.from_str(
        (f"{inspect.getsource(kubernetes_enabled)}" "\nkubernetes_enabled()"),
        ["kubernetes"],
        [
            ImportFromModule(module="kubernetes.client", imports=("CoreV1Api",)),
            ImportFromModule(module="kubernetes.config", imports=("load_config",)),
        ],
    )
    async with PodCommandLineCodeExecutor(functions=[test_function]) as executor:
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code="print('Hello, World!')"),
            ],
            cancellation_token=CancellationToken(),
        )
        assert re.fullmatch(generated_pod_name_regex, executor._pod_name) is not None
        assert code_result.exit_code == 0
        assert "Hello, World!" in code_result.output


@with_requirements(
    python_packages=["httpx"],
    global_imports=[
        Alias(name="httpx", alias="hx"),
        ImportFromModule(module="typing", imports=("Any",)),
    ],
)
def load_data() -> Any:
    """
    fetch cat fact api

    Returns:
        Any: status for request to cat fact api
    """
    client = hx.Client()
    if client.base_url is None:
        pass
    return "success"


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_func_modules_with_requirements(generated_pod_name_regex: str) -> None:
    async with PodCommandLineCodeExecutor(functions=[load_data]) as executor:
        code = f"from {executor._functions_module} import load_data;print(load_data())"
        code_result = await executor.execute_code_blocks(
            code_blocks=[
                CodeBlock(language="python", code=code),
            ],
            cancellation_token=CancellationToken(),
        )
        assert re.fullmatch(generated_pod_name_regex, executor._pod_name) is not None
        assert code_result.exit_code == 0
        assert "success" in code_result.output


@pytest.mark.skipif(not state_kubernetes_enabled, reason="kubernetes not accessible")
@pytest.mark.asyncio
async def test_with_code_executor_agent() -> None:
    async with PodCommandLineCodeExecutor() as executor:
        code_executor_agent = CodeExecutorAgent("code_executor", code_executor=executor)
        task = TextMessage(
            content="""Here is some code
```python
print('Hello world!')
```
""",
            source="user",
        )
        response = await code_executor_agent.on_messages([task], CancellationToken())
        assert "Hello world!" in response.chat_message.content
